import BackNumpy as B
import math
import numpy as np
import time

class PredictionFeedbackESN:
    def __init__(self, ruido, fuga, contenedor, semillaW, semillaP, densidad, bias, entrada, salidas, store=0 ):
        '''
            ruido       [0,1]
            fuga        [0,1]
            contenedor  [5,1000]
            semillaW    [0,100]
            semillaP    [0,100]
            densidad    [0.1,1]
        '''
        self.ruido = ruido
        self.fuga = fuga
        self.contenedor = contenedor
        self.semillaW = semillaW
        self.semillaP = semillaP
        self.densidad = densidad
        self.bias = bias
        self.entrada = entrada
        self.salidas = salidas
        self.outputBias=1.0,
        self.outputInputScaling=1.0
        self.feedbackScaling=1.0,
        self.activation = B.TanH
        #self.activation = B.activation_r
        self.activation_relu = B.activation_r
        self.out_inverse_activation = lambda x: x
        self.out_inverse_activation_relu =  B.inv_activation_r
        self.out_activation = lambda x: x
        
        self.matriz_entrada( store )
        self.matriz_contenedor( store )
        self.matriz_retroalimentacion()
    
    def fitCompacta(self, DatosX, DatosY, store=0 ):

        if(DatosX.shape[0] != DatosY.shape[0]):
            raise ValueError("Las dimensiones de los datos de entrada y salida no coinciden.")

        muestras = DatosX.shape[0]

        self._X = B.Empty((1 + self.entrada + self.contenedor, muestras))
        self.xmin = B.Zeros((self.contenedor, 1))
        Y_target = B.Empty((self.salidas, muestras))

        for i in range(muestras):
            u = DatosX[i].reshape(self.entrada, 1)
            T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
            self.xmin = self.xmin * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
            self._X[:,i] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin) )[:, 0]
            Y_target[:,i] = self.out_inverse_activation(DatosY[i]).T

        self.Wout = B.Dot(Y_target, B.Pinv(self._X))
        if store == 1 :
            np.savetxt( "Wout.txt", self.Wout )

    
    def matriz_entrada(self):
        B.Seed(self.semillaW) 
        self.Win = B.Random(self.contenedor, 1 + self.entrada) - 0.5

    def matriz_retroalimentacion(self):
        B.Seed(self.semillaW)
        self.Wfeed = B.Random(self.contenedor, 1 + self.salidas) - 0.5
        
    def matriz_contenedor(self):
        self.permutacion()
        self.calculardensidad()
        Grupos = math.ceil(self.celdas / self.contenedor)
        B.Seed(self.semillaW)
        diagonal=B.Random(self.contenedor, Grupos)-0.5

        self.W = B.Zeros((self.contenedor, self.contenedor))
        #self.Ws = B.Zeros((self.contenedor, 2), dtype=int)
        for i in range(self.contenedor):
            for j in range(Grupos):
                encaje = self.p[i] + j
                if encaje > (self.contenedor - 1):
                    encaje = encaje - self.contenedor
        
                self.W[i,encaje] = diagonal[self.p[i], j]
                #self.Ws[i,0] = i
                #self.Ws[i,1] = encaje

        #print(np.linalg.eig(self.W)[0])
        Weigen =B.Abs( B.Eigen(self.W)[0] )
        #print(Weigen)
        self.W *= 1 / B.Max(Weigen)
        '''
        for i in range(self.contenedor):
            self.Ws[i,0] = self.W[int(self.Ws[i,1]), int(self.Ws[i,2])]
        '''

    def permutacion(self):
        B.Seed(self.semillaP)
        self.p = B.Permutation(self.contenedor)
    
    def calculardensidad(self):
        # Se calcula el número de celdas que se van hacer diferentes de 0
        # f(densidad) = celdas
        # celdas = A*densidad + B*densidad + C
        n2 = self.contenedor * self.contenedor
        n = self.contenedor
        a = -5/18*n2 + 25/9*n
        b = 17/12*n2 - 25/6*n
        c = -5/36*n2 + 25/18*n
        self.celdas = math.floor( a*(self.densidad*self.densidad) + b*self.densidad + c)

class PredictionDeepESN:
    def __init__(self, ruido, fuga, contenedor, semillaW, semillaP, densidad, bias, entrada, salidas):
        '''
            ruido       [0,1]
            fuga        [0,1]
            contenedor  [5,1000]
            semillaW    [0,100]
            semillaP    [0,100]
            densidad    [0.1,1]
        '''
        self.ruido = ruido
        self.fuga = fuga
        self.contenedor = contenedor
        self.semillaW = semillaW
        self.semillaP = semillaP
        self.densidad = densidad
        self.bias = bias
        self.entrada = entrada
        self.salidas = salidas
        self.outputBias=1.0,
        self.outputInputScaling=1.0
        self.feedbackScaling=1.0,
        self.activation = B.TanH
        #self.activation = B.activation_r
        self.activation_relu = B.activation_r
        self.out_inverse_activation = lambda x: x
        self.out_inverse_activation_relu =  B.inv_activation_r
        self.out_activation = lambda x: x
        
        self.matriz_entrada()
        self.matriz_contenedor()

    def fitCompacta(self, DatosX, DatosY):

        if(DatosX.shape[0] != DatosY.shape[0]):
            raise ValueError("Las dimensiones de los datos de entrada y salida no coinciden.")

        muestras = DatosX.shape[0]

        self._X = B.Empty((1 + self.entrada + self.contenedor, muestras))
        self.xmin = B.Zeros((self.contenedor, 1))
        self.xmin2 = B.Zeros((self.contenedor, 1))
        Y_target = B.Empty((self.salidas, muestras))
       
     
        for i in range(muestras):
            u = DatosX[i].reshape(self.entrada, 1)
            T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
            self.xmin = self.xmin * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
            
            T2 = B.Dot(self.Win2, T) + B.Dot(self.W, self.xmin2)
            self.xmin2 = self.xmin2 * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin2 += self.fuga * self.activation(T2 + (B.Random() - 0.5) * self.ruido)


            self._X[:,i] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin2) )[:, 0]
            Y_target[:,i] = self.out_inverse_activation(DatosY[i]).T

        self.Wout = B.Dot(Y_target, B.Pinv(self._X))

        #print(self.Wout.shape)

    def predictCompacta(self, DatosX):
        muestras = DatosX.shape[0]
        continuacion = False
        if not continuacion:
            self.xmin = B.Zeros((self.contenedor, 1))
            self.xmin2 = B.Zeros((self.contenedor, 1))

        #Se debe de guardar también el vector de retroalimentación
        self._X = B.Empty((1 + self.entrada + self.contenedor, muestras))
        for i in range(muestras):
            u = DatosX[i].reshape(self.entrada, 1)
            T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
            self.xmin = self.xmin * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
            
            T2 = B.Dot(self.Win2, T) + B.Dot(self.W, self.xmin2)
            self.xmin2 = self.xmin2 * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin2 += self.fuga * self.activation(T2 + (B.Random() - 0.5) * self.ruido)


            self._X[:,i] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin2) )[:, 0]
        

        Y = B.Dot(self.Wout, self._X)

        return Y.T
     
    def matriz_entrada(self):
        B.Seed(self.semillaW) 
        self.Win = B.Random(self.contenedor, 1 + self.entrada) - 0.5
        self.Win2 = B.Random(self.contenedor, self.contenedor) - 0.5
        
    def matriz_contenedor(self):
        self.permutacion()
        self.calculardensidad()
        Grupos = math.ceil(self.celdas / self.contenedor)
        B.Seed(self.semillaW)
        diagonal=B.Random(self.contenedor, Grupos)-0.5

        self.W = B.Zeros((self.contenedor, self.contenedor))
        #self.Ws = B.Zeros((self.contenedor, 2), dtype=int)
        for i in range(self.contenedor):
            for j in range(Grupos):
                encaje = self.p[i] + j
                if encaje > (self.contenedor - 1):
                    encaje = encaje - self.contenedor
        
                self.W[i,encaje] = diagonal[self.p[i], j]
                #self.Ws[i,0] = i
                #self.Ws[i,1] = encaje

        #print(np.linalg.eig(self.W)[0])
        Weigen =B.Abs( B.Eigen(self.W)[0] )
        #print(Weigen)
        self.W *= 1 / B.Max(Weigen)
        #self.W2 = self.W
        '''
        for i in range(self.contenedor):
            self.Ws[i,0] = self.W[int(self.Ws[i,1]), int(self.Ws[i,2])]
        '''

    def permutacion(self):
        B.Seed(self.semillaP)
        self.p = B.Permutation(self.contenedor)
    
    def calculardensidad(self):
        # Se calcula el número de celdas que se van hacer diferentes de 0
        # f(densidad) = celdas
        # celdas = A*densidad + B*densidad + C
        n2 = self.contenedor * self.contenedor
        n = self.contenedor
        a = -5/18*n2 + 25/9*n
        b = 17/12*n2 - 25/6*n
        c = -5/36*n2 + 25/18*n
        self.celdas = math.floor( a*(self.densidad*self.densidad) + b*self.densidad + c)

class PredictionESN:
    def __init__(self, ruido, fuga, contenedor, semillaW, semillaP, densidad, bias, entrada, salidas):
        '''
            ruido       [0,1]
            fuga        [0,1]
            contenedor  [5,1000]
            semillaW    [0,100]
            semillaP    [0,100]
            densidad    [0.1,1]
        '''
        self.ruido = ruido
        self.fuga = fuga
        self.contenedor = contenedor
        self.semillaW = semillaW
        self.semillaP = semillaP
        self.densidad = densidad
        self.bias = bias
        self.entrada = entrada
        self.salidas = salidas
        self.outputBias=1.0,
        self.outputInputScaling=1.0
        self.feedbackScaling=1.0,
        self.activation = B.TanH
        #self.activation = B.activation_r
        self.activation_relu = B.activation_r
        self.out_inverse_activation = lambda x: x
        self.out_inverse_activation_relu =  B.inv_activation_r
        self.out_activation = lambda x: x
        
        self.matriz_entrada()
        self.matriz_contenedor()

    def fitCompacta(self, DatosX, DatosY):

        if(DatosX.shape[0] != DatosY.shape[0]):
            raise ValueError("Las dimensiones de los datos de entrada y salida no coinciden.")

        muestras = DatosX.shape[0]

        self._X = B.Empty((1 + self.entrada + self.contenedor, muestras))
        self.xmin = B.Zeros((self.contenedor, 1))
        Y_target = B.Empty((self.salidas, muestras))
       
     
        for i in range(muestras):
            u = DatosX[i].reshape(self.entrada, 1)
            T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
            self.xmin = self.xmin * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
            self._X[:,i] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin) )[:, 0]
            Y_target[:,i] = self.out_inverse_activation(DatosY[i]).T

        self.Wout = B.Dot(Y_target, B.Pinv(self._X))


    def predictCompacta(self, DatosX):
        muestras = DatosX.shape[0]
        continuacion = False
        if not continuacion:
            self.xmin = B.Zeros((self.contenedor, 1))

        #Se debe de guardar también el vector de retroalimentación
        self._X = B.Empty((1 + self.entrada + self.contenedor, muestras))
        for i in range(muestras):
            u = DatosX[i].reshape(self.entrada, 1)
            T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
            self.xmin = self.xmin * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
            self._X[:,i] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin) )[:, 0]
        

        Y = B.Dot(self.Wout, self._X)

        return Y.T
     
    def matriz_entrada(self):
        B.Seed(self.semillaW) 
        self.Win = B.Random(self.contenedor, 1 + self.entrada) - 0.5
        
    def matriz_contenedor(self):
        self.permutacion()
        self.calculardensidad()
        Grupos = math.ceil(self.celdas / self.contenedor)
        B.Seed(self.semillaW)
        diagonal=B.Random(self.contenedor, Grupos)-0.5

        self.W = B.Zeros((self.contenedor, self.contenedor))
        #self.Ws = B.Zeros((self.contenedor, 2), dtype=int)
        for i in range(self.contenedor):
            for j in range(Grupos):
                encaje = self.p[i] + j
                if encaje > (self.contenedor - 1):
                    encaje = encaje - self.contenedor
        
                self.W[i,encaje] = diagonal[self.p[i], j]
                #self.Ws[i,0] = i
                #self.Ws[i,1] = encaje

        #print(np.linalg.eig(self.W)[0])
        Weigen =B.Abs( B.Eigen(self.W)[0] )
        #print(Weigen)
        self.W *= 1 / B.Max(Weigen)
        '''
        for i in range(self.contenedor):
            self.Ws[i,0] = self.W[int(self.Ws[i,1]), int(self.Ws[i,2])]
        '''

    def permutacion(self):
        B.Seed(self.semillaP)
        self.p = B.Permutation(self.contenedor)
    
    def calculardensidad(self):
        # Se calcula el número de celdas que se van hacer diferentes de 0
        # f(densidad) = celdas
        # celdas = A*densidad + B*densidad + C
        n2 = self.contenedor * self.contenedor
        n = self.contenedor
        a = -5/18*n2 + 25/9*n
        b = 17/12*n2 - 25/6*n
        c = -5/36*n2 + 25/18*n
        self.celdas = math.floor( a*(self.densidad*self.densidad) + b*self.densidad + c)

class PredictionIncESN:
    def __init__(self, ruido, fuga, contenedor, semillaW, semillaP, densidad, bias, entrada, salidas):
        '''
            ruido       [0,1]
            fuga        [0,1]
            contenedor  [5,1000]
            semillaW    [0,100]
            semillaP    [0,100]
            densidad    [0.1,1]
        '''
        self.ruido = ruido
        self.fuga = fuga
        self.contenedor = contenedor
        self.semillaW = semillaW
        self.semillaP = semillaP
        self.densidad = densidad
        self.bias = bias
        self.entrada = entrada
        self.salidas = salidas
        self.outputBias=1.0,
        self.outputInputScaling=1.0
        self.feedbackScaling=1.0,
        self.activation = B.TanH
        #self.activation = B.activation_r
        self.activation_relu = B.activation_r
        self.out_inverse_activation = lambda x: x
        self.out_inverse_activation_relu =  B.inv_activation_r
        self.out_activation = lambda x: x
        
        self.matriz_entrada()
        self.matriz_contenedor()

    def fitCompacta(self, DatosX, DatosY):

        if(DatosX.shape[0] != DatosY.shape[0]):
            raise ValueError("Las dimensiones de los datos de entrada y salida no coinciden.")

        muestras = DatosX.shape[0]

        self._X = B.Empty((1 + self.entrada + self.contenedor, muestras))
        self.xmin = B.Zeros((self.contenedor, 1))
        Y_target = B.Empty((self.salidas, muestras))
       
     
        for i in range(muestras):
            u = DatosX[i].reshape(self.entrada, 1)
            T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
            self.xmin = self.xmin * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
            self._X[:,i] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin) )[:, 0]
            Y_target[:,i] = self.out_inverse_activation(DatosY[i]).T
        
        self.Wout = B.Dot(Y_target, B.Pinv(self._X))
        

    def fitCompactaIncremental(self, DatosX, DatosY, incializar_estado=True):

        if(DatosX.shape[0] != DatosY.shape[0]):
            raise ValueError("Las dimensiones de los datos de entrada y salida no coinciden.")

        muestras = DatosX.shape[0]
        over = 0
        #En caso de que ya se haya inicializado, no se vuelve a resetear el estado anterior
        if incializar_estado:
            self.xmin = B.Zeros((self.contenedor, 1))
        
        #print("x estados")
        #print(self.xmin)

        self._X = B.Empty((1 + self.entrada + self.contenedor, muestras))
        
        for i in range(muestras):
            u = DatosX[i].reshape(self.entrada, 1)
            T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
            self.xmin = self.xmin * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
            self._X[:,i] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin) )[:, 0]
        
        b = DatosY.reshape((DatosY.shape[0], 1))
        A = np.transpose(self._X)
        #print(np.linalg.matrix_rank(A))
        n1, m = A.shape
        
        # Esta parte unícamente se va a realizar para inicializar la matriz una vez inicializada 
        # Ya se utilizará la matriz precalculada

        if incializar_estado:
            n = m + 500
            C = A[0:n, :]
            #self.H = np.linalg.inv( np.transpose(C) @ C )
            self.H = B.Pinv( np.transpose(C) @ C )
            self.x = self.H @ ( np.transpose(C) @ b[0:n])
            i = n
        else:
            i = 0
        #print("x wout")
        #print(self.x)
        
        while i < n1: 
            a = A[i]
            c = a.reshape((m,1))
            s = self.H @ c
            self.H = self.H - (s @ s.T)/(1.0 + c.T @ s )
            v = self.H @ c
    
            self.x = self.x + v*( b[i].reshape((1,1)) - c.T @ self.x )
            tmp = np.sum(self.x)
            if (np.isnan(tmp) or np.isinf(tmp))  :
                #print("NAN encontrado")
                over = 1
                break
            i += 1
        #print(x)
        if over == 1:
            self.WoutInc = 1
        else:
            self.WoutInc = np.transpose(self.x)

    def predictCompacta(self, DatosX):
        muestras = DatosX.shape[0]
        continuacion = False
        if not continuacion:
            self.xmin = B.Zeros((self.contenedor, 1))

        #Se debe de guardar también el vector de retroalimentación
        self._X = B.Empty((1 + self.entrada + self.contenedor, muestras))
        for i in range(muestras):
            u = DatosX[i].reshape(self.entrada, 1)
            T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
            self.xmin = self.xmin * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
            self._X[:,i] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin) )[:, 0]
        

        Y = B.Dot(self.Wout, self._X)

        return Y.T

    def predictIncCompacta(self, DatosX):
        muestras = DatosX.shape[0]
        continuacion = False
        if not continuacion:
            self.xmin = B.Zeros((self.contenedor, 1))

        #Se debe de guardar también el vector de retroalimentación
        self._X = B.Empty((1 + self.entrada + self.contenedor, muestras))
        for i in range(muestras):
            u = DatosX[i].reshape(self.entrada, 1)
            T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
            self.xmin = self.xmin * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
            self._X[:,i] = B.Vs( ( B.Array(self.outputBias), self.outputInputScaling * u, self.xmin) )[:, 0]
        

        Y = B.Dot(self.WoutInc, self._X)

        return Y.T    
     
    def matriz_entrada(self):
        B.Seed(self.semillaW) 
        self.Win = B.Random(self.contenedor, 1 + self.entrada) - 0.5
        
    def matriz_contenedor(self):
        self.permutacion()
        self.calculardensidad()
        Grupos = math.ceil(self.celdas / self.contenedor)
        B.Seed(self.semillaW)
        diagonal=B.Random(self.contenedor, Grupos)-0.5

        self.W = B.Zeros((self.contenedor, self.contenedor))
        #self.Ws = B.Zeros((self.contenedor, 2), dtype=int)
        for i in range(self.contenedor):
            for j in range(Grupos):
                encaje = self.p[i] + j
                if encaje > (self.contenedor - 1):
                    encaje = encaje - self.contenedor
        
                self.W[i,encaje] = diagonal[self.p[i], j]
                #self.Ws[i,0] = i
                #self.Ws[i,1] = encaje

        #print(np.linalg.eig(self.W)[0])
        Weigen =B.Abs( B.Eigen(self.W)[0] )
        #print(Weigen)
        self.W *= 1 / B.Max(Weigen)
        '''
        for i in range(self.contenedor):
            self.Ws[i,0] = self.W[int(self.Ws[i,1]), int(self.Ws[i,2])]
        '''

    def permutacion(self):
        B.Seed(self.semillaP)
        self.p = B.Permutation(self.contenedor)
    
    def calculardensidad(self):
        # Se calcula el número de celdas que se van hacer diferentes de 0
        # f(densidad) = celdas
        # celdas = A*densidad + B*densidad + C
        n2 = self.contenedor * self.contenedor
        n = self.contenedor
        a = -5/18*n2 + 25/9*n
        b = 17/12*n2 - 25/6*n
        c = -5/36*n2 + 25/18*n
        self.celdas = math.floor( a*(self.densidad*self.densidad) + b*self.densidad + c)

class PredictionEncoderESN:
    def __init__(self, ruido, fuga, contenedor, semillaW, semillaP, densidad, bias, entrada, salidas):
        '''
            ruido       [0,1]
            fuga        [0,1]
            contenedor  [5,1000]
            semillaW    [0,100]
            semillaP    [0,100]
            densidad    [0.1,1]
        '''
        self.ruido = ruido
        self.fuga = fuga
        self.contenedor = contenedor
        self.semillaW = semillaW
        self.semillaP = semillaP
        self.densidad = densidad
        self.bias = bias
        self.entrada = entrada
        self.salidas = salidas
        self.outputBias=1.0,
        self.outputInputScaling=1.0
        self.feedbackScaling=1.0,
        #self.activation = B.TanH
        self.activation = B.activation_r
        self.activation_relu = B.activation_r
        self.out_inverse_activation = lambda x: x
        self.out_inverse_activation_relu =  B.inv_activation_r
        self.out_activation = lambda x: x
        
        self.matriz_entrada()
        self.matriz_contenedor()
        self.matriz_salida()

    def fitCompacta(self, DatosX, DatosY):

        if(DatosX.shape[0] != DatosY.shape[0]):
            raise ValueError("Las dimensiones de los datos de entrada y salida no coinciden.")

        muestras = DatosX.shape[0]

        self._X = B.Empty((1 + self.entrada + self.contenedor, muestras))
        self.xmin = B.Zeros((self.contenedor, 1))
        Y_target = B.Empty((self.salidas, muestras))
       
     
        for i in range(muestras):
            u = DatosX[i].reshape(self.entrada, 1)
            T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
            self.xmin = self.xmin * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
            self._X[:,i] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin) )[:, 0]
            Y_target[:,i] = self.out_inverse_activation(DatosY[i]).T

        
        self.Wout = B.Dot(Y_target, B.Pinv(self._X))

        #print(self.Wout.shape)

    def predictCompacta(self, DatosX):
        muestras = DatosX.shape[0]
        continuacion = False
        if not continuacion:
            self.xmin = B.Zeros((self.contenedor, 1))

        #Se debe de guardar también el vector de retroalimentación
        self._X = B.Empty((1 + self.entrada + self.contenedor, muestras))
        for i in range(muestras):
            u = DatosX[i].reshape(self.entrada, 1)
            T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
            self.xmin = self.xmin * (1 - self.fuga)
            B.Seed(self.semillaW)
            self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
            self._X[:,i] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin) )[:, 0]
        Y = B.Dot(self.Wout, self._X)

        return Y.T
     
    def matriz_entrada(self):
        B.Seed(self.semillaW) 
        self.Win = B.Random(self.contenedor, 1 + self.entrada) - 0.5

    def matriz_salida(self):
        B.Seed(self.semillaW) 
        self.Wout = B.Random(self.salidas, 1 + self.entrada + self.contenedor) - 0.5
        
    def matriz_contenedor(self):
        self.permutacion()
        self.calculardensidad()
        Grupos = math.ceil(self.celdas / self.contenedor)
        B.Seed(self.semillaW)
        diagonal=B.Random(self.contenedor, Grupos)-0.5

        self.W = B.Zeros((self.contenedor, self.contenedor))
        #self.Ws = B.Zeros((self.contenedor, 2), dtype=int)
        for i in range(self.contenedor):
            for j in range(Grupos):
                encaje = self.p[i] + j
                if encaje > (self.contenedor - 1):
                    encaje = encaje - self.contenedor
        
                self.W[i,encaje] = diagonal[self.p[i], j]
                #self.Ws[i,0] = i
                #self.Ws[i,1] = encaje

        #print(np.linalg.eig(self.W)[0])
        Weigen =B.Abs( B.Eigen(self.W)[0] )
        #print(Weigen)
        self.W *= 1 / B.Max(Weigen)
        '''
        for i in range(self.contenedor):
            self.Ws[i,0] = self.W[int(self.Ws[i,1]), int(self.Ws[i,2])]
        '''

    def permutacion(self):
        B.Seed(self.semillaP)
        self.p = B.Permutation(self.contenedor)
    
    def calculardensidad(self):
        # Se calcula el número de celdas que se van hacer diferentes de 0
        # f(densidad) = celdas
        # celdas = A*densidad + B*densidad + C
        n2 = self.contenedor * self.contenedor
        n = self.contenedor
        a = -5/18*n2 + 25/9*n
        b = 17/12*n2 - 25/6*n
        c = -5/36*n2 + 25/18*n
        self.celdas = math.floor( a*(self.densidad*self.densidad) + b*self.densidad + c)


class ClasificacionESN:
    def __init__(self, ruido, fuga, contenedor, semillaW, semillaP, densidad, bias, entrada, clases, store=0):
        '''
            ruido       [0,1]
            fuga        [0,1]
            contenedor  [5,1000]
            semillaW    [0,100]
            semillaP    [0,100]
            densidad    [0.1,1]
        '''

        self.ruido = ruido
        self.fuga = fuga
        self.contenedor = contenedor
        self.semillaW = semillaW
        self.semillaP = semillaP
        self.densidad = densidad
        self.bias = bias
        self.entrada = entrada
        self.clases = clases
        self.outputBias=1.0,
        self.outputInputScaling=1.0
        self.activation = B.TanH
        self.activation_relu = B.activation_r
        self.out_inverse_activation = lambda x: B.Log((x * 0.98 + 0.01) / (0.99 - x * 0.98))
        self.out_inverse_activation_relu =  B.inv_activation_r
        self.out_activation = lambda x: 0.1 + 0.98 * x / (1 + B.Exp(-x))
        
        self.matriz_entrada(store)
        self.matriz_contenedor(store)

    def loadMatrices( self, win, per, w, wout ) :
        self.Win = np.loadtxt( win )
        # print( "Win:", self.Win.shape )
        self.Wout = np.loadtxt( wout )
        diagonal= np.loadtxt( w )
        diagonal = diagonal.reshape( -1, 1 )
        self.p = np.loadtxt( per, dtype=int )

        self.calculardensidad()
        Grupos = math.ceil(self.celdas / self.contenedor)
        # print( "Grupos 2:", Grupos )
        self.W = B.Zeros((self.contenedor, self.contenedor))
        #self.Ws = B.Zeros((self.contenedor, 2), dtype=int)
        for i in range(self.contenedor):
            for j in range(Grupos):
                encaje = self.p[i] + j
                if encaje > (self.contenedor - 1):
                    encaje = encaje - self.contenedor
        
                self.W[i,encaje] = diagonal[self.p[i], j]

    def fitCompacta(self, DatosX, DatosY, store=0):
        if(DatosX.shape[0] != DatosY.shape[0]):
            raise ValueError("Las dimensiones de los datos de entrada y salida no coinciden.")
        
        muestras = DatosX.shape[0]
        caracteristicas = DatosX.shape[1]

        if(DatosY.shape[1] > 1 and DatosY.shape[1] != self.clases):
            raise ValueError("El número de clases en los datos de salida no coinciden con el número de clases.")
        
        self.xmin = B.Zeros((self.contenedor, 1))

        self.Xmay = B.Zeros((
            1+self.entrada+self.contenedor, 
            (muestras*caracteristicas)
        ))

        Y_objetivo = B.Zeros((
            self.clases,(muestras*caracteristicas)
        ))

        for i in range(muestras):
            self.xmin = B.Zeros((self.contenedor, 1))
            for j in range(caracteristicas):
                u = DatosX[i,j]
                T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
                self.xmin = self.xmin * (1 - self.fuga)
                B.Seed(self.semillaW)
                #self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
                self.xmin += self.fuga * self.activation_relu(T + (B.Random() - 0.5) * self.ruido)
                self.Xmay[:,i*caracteristicas+j] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin)  )[:, 0]
                #Y_objetivo[:,i*caracteristicas+j] = self.out_inverse_activation(DatosY[i]).T
                Y_objetivo[:,i*caracteristicas+j] = self.out_inverse_activation_relu(DatosY[i]).T
                
                
        self.Wout = B.Dot(Y_objetivo, B.Pinv(self.Xmay))
        if store == 1 :
            np.savetxt( "Wout.txt", self.Wout )

    def predictCompacta(self, DatosX):
        caracteristicas = DatosX.shape[1]
        muestras = DatosX.shape[0]
        
        Y = B.Empty((DatosX.shape[0], self.clases))
        self.Xmay = B.Zeros((
            1+self.entrada+self.contenedor, caracteristicas
        ))

        for i in range(muestras):
            self.xmin = B.Zeros((self.contenedor, 1))
            for j in range(caracteristicas):
                u = DatosX[i,j]
                T = B.Dot(self.Win, B.Vs((B.Array(self.bias), u))) + B.Dot(self.W, self.xmin)
                self.xmin = self.xmin * (1 - self.fuga)
                B.Seed(self.semillaW)
                #self.xmin += self.fuga * self.activation(T + (B.Random() - 0.5) * self.ruido)
                self.xmin += self.fuga * self.activation_relu(T + (B.Random() - 0.5) * self.ruido)
                self.Xmay[:,j] = B.Vs( (B.Array(self.outputBias), self.outputInputScaling * u, self.xmin)  )[:, 0]
            
            y = B.Dot(self.Wout, self.Xmay)
            
            Y[i] = B.Mean(y,1)
        
        resultado = B.Zeros(Y.shape)
        
        for i, n in enumerate(B.ArgMax(Y,1)):
            resultado[i, n] = 1.0
   
        return resultado

    def mult_dispersa(self, x):
        R = B.Zeros((x.shape))
        for i in range(len(self.W)):
            R[i] = self.W[i,self.Ws[i,1]] * x[self.Ws[i,1],0]
        return R
     

    def matriz_entrada(self, store=0):
        B.Seed(self.semillaW) 
        self.Win = B.Random(self.contenedor, 2) - 0.5
        if store==1 :
            np.savetxt( "Win.txt", self.Win )
        
    def matriz_contenedor(self, store=0):
        self.permutacion()
        self.calculardensidad()
        Grupos = math.ceil(self.celdas / self.contenedor)
        # print( "Grupos:", Grupos )
        B.Seed(self.semillaW)
        diagonal=B.Random(self.contenedor, Grupos)-0.5

        self.W = B.Zeros((self.contenedor, self.contenedor))
        #self.Ws = B.Zeros((self.contenedor, 2), dtype=int)
        for i in range(self.contenedor):
            for j in range(Grupos):
                encaje = self.p[i] + j
                if encaje > (self.contenedor - 1):
                    encaje = encaje - self.contenedor
        
                self.W[i,encaje] = diagonal[self.p[i], j]
                #self.Ws[i,0] = i
                #self.Ws[i,1] = encaje

        #print(np.linalg.eig(self.W)[0])
        Weigen =B.Abs( B.Eigen(self.W)[0] )
        #print(Weigen)
		# HERE
        self.W *= 1 / B.Max(Weigen)
        if store==1 :
            np.savetxt( "permutation.txt", self.p )
            np.savetxt( "w.txt", diagonal/B.Max(Weigen) )
        '''
        for i in range(self.contenedor):
            self.Ws[i,0] = self.W[int(self.Ws[i,1]), int(self.Ws[i,2])]
        '''


    def permutacion(self):
        B.Seed(self.semillaP)
        self.p = B.Permutation(self.contenedor)
    
    def calculardensidad(self):
        # Se calcula el número de celdas que se van hacer diferentes de 0
        # f(densidad) = celdas
        # celdas = A*densidad + B*densidad + C
        n2 = self.contenedor * self.contenedor
        n = self.contenedor
        a = -5/18*n2 + 25/9*n
        b = 17/12*n2 - 25/6*n
        c = -5/36*n2 + 25/18*n
        self.celdas = math.floor( a*(self.densidad*self.densidad) + b*self.densidad + c)
